import unittest
from zlepy.fgen import get_po, rebuildlists

class TestFgenGen(unittest.TestCase):
    def test_get_po(self):
        result = get_po([5])
        self.assertIsInstance(result, list)

    def test_rebuildlists(self):
        result = rebuildlists(3)
        self.assertIsInstance(result, list)

    # TODO add test for genmat


if __name__ == '__main__':
    unittest.main()


import numpy as np
import itertools
from operator import itemgetter
from collections import defaultdict, deque
import sympy as sp

def argsort(seq):
    return sorted(range(len(seq)), key=seq.__getitem__)

def perm_inverse(perm):
    inv = np.empty_like(perm)
    inv[perm] = np.arange(len(inv), dtype=inv.dtype)
    return inv

#can speed-up by taking the fastest algorithm for finding disconnected components
def po_groups(po):
    groups = [po[0:2]] 
    for i in range(1,int(len(po)/2)):
        relation = po[2*i:2*i+2]
        booli = 1
        #can map isin to avoid looping
        for j in range(len(groups)):
            if (np.any(np.in1d(relation,groups[j]))): #check if any node in relation belongs to given group j
                groups[j]=np.append(groups[j], tuple(relation)) #add to group if so, then break
                booli = 0
                break
        if (booli):
            groups.append(relation)
    return groups

def edges_to_adjacency_list(edges):
    adjacency_list = defaultdict(list)
    for i in range(0, len(edges), 2):
        u = edges[i]
        v = edges[i + 1]
        adjacency_list[u].append(v)
    
    # Ensure all nodes are included, even if they have no outgoing edges
    nodes = set(edges)
    for node in nodes:
        if node not in adjacency_list:
            adjacency_list[node] = []

    return adjacency_list

def merge_components(comps):
    merged_dict = defaultdict(list)
    for d in comps:
        merged_dict.update(d)
    return merged_dict

def add_missing_nodes(edge_list):
    all_nodes = set(edge_list.keys())
    
    # Include nodes that are only in the values
    for nodes in edge_list.values():
        all_nodes.update(nodes)
    
    # Ensure every node has a key in the dictionary
    for node in all_nodes:
        if node not in edge_list:
            edge_list[node] = []
    
    return edge_list


#transitive reduction
def topological_sort(graph):
    in_degree = defaultdict(int)
    for node in graph:
        for neighbor in graph[node]:
            in_degree[neighbor] += 1
    
    zero_in_degree_queue = deque([node for node in graph if in_degree[node] == 0])
    topo_order = []
    
    count = 0
    while zero_in_degree_queue:
        node = zero_in_degree_queue.popleft()
        topo_order.append(node)
        count += 1
        for neighbor in graph[node]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                zero_in_degree_queue.append(neighbor)
    
    return topo_order

def longest_path_dag(graph):
    topo_order = topological_sort(graph)
    dist = {node: float('-inf') for node in graph}
    
    for node in graph:
        if node not in dist or dist[node] == float('-inf'):
            dist[node] = 0
    
    for node in topo_order:
        for neighbor in graph[node]:
            if dist[neighbor] < dist[node] + 1:
                dist[neighbor] = dist[node] + 1
    
    return dist

def transitive_reduction(graph):
    topo_order = topological_sort(graph)
    longest_paths = longest_path_dag(graph)
    reduced_graph = defaultdict(list)
    
    # collect longest paths
    for node in topo_order:
        for neighbor in graph[node]:
            if longest_paths[neighbor] == longest_paths[node] + 1:
                reduced_graph[node].append(neighbor)
    
    # ensure all nodes are preserved and maintain reachability
    for node in graph:
        if node not in reduced_graph:
            reduced_graph[node] = []
        for neighbor in graph[node]:
            if neighbor not in reduced_graph[node]:
                reduced_graph[node].append(neighbor)

    # remove shorter paths by checking if any node has more than one parent in the reduced graph
    for node in graph:
        for neighbor in graph[node]:
            for descendant in reduced_graph[neighbor]:
                if descendant in reduced_graph[node]:
                    reduced_graph[node].remove(descendant)
    
    return reduced_graph

#levels / by-variable permutation restriction
def level_sort(graph):        
    in_degree = defaultdict(int)
    for node in graph:
        for neighbor in graph[node]:
            in_degree[neighbor] += 1
    
    zero_in_degree_queue = deque([node for node in graph if in_degree[node] == 0])
    zero_curr = [node for node in graph if in_degree[node] == 0]

    count = 0
    levels = defaultdict(list)
    levels[count] = zero_curr

    while zero_in_degree_queue:
        if (not any(np.isin(zero_curr, zero_in_degree_queue))):
            zero_curr = list(set([node for node in graph if in_degree[node] == 0])-set([x for xs in list(levels.values()) for x in xs]))
            count += 1
            levels[count] = zero_curr

        node = zero_in_degree_queue.popleft()
        for neighbor in graph[node]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                zero_in_degree_queue.append(neighbor)

    return levels

def invert_dict_of_lists(d):
    inverted = defaultdict(list)
    for key, values in d.items():
        for value in values:
            inverted[value] = key
    return inverted

def reverse_graph(graph):
    reversed_graph = defaultdict(list)
    for source, targets in graph.items():
        for target in targets:
            reversed_graph[target].append(source)
    return reversed_graph


#block-variable construction
def compute_degrees(graph, node):
    in_degree = 0
    out_degree = len(graph[node])
    
    #returns in/out-degree in order of keys present.
    for key in graph:
        if node in graph[key]:
            in_degree += 1
            
    return [in_degree, out_degree]

def find_max_chain(start, graph, in_out):
    chain = []
    #start node
    chain.append(start)
    if (in_out[start][1] == 1):
        current = graph[start][0]
        #intermediate nodes
        while current in graph and in_out[current][0] == 1 and in_out[current][1] == 1:
            chain.append(current)
            current = graph[current][0]
        #final node
        if (in_out[current][0] == 1):
            chain.append(current)
    return chain

def in_out_dict(graph):   
    in_out = defaultdict(list, {key: [0, 0] for key in list(graph.keys())})
    for node in graph:
        in_out[node][1] = len(graph[node])
        for neighbor in graph[node]:
            in_out[neighbor][0] += 1
            
    return in_out

def contract_graph(graph):
    block_graph = defaultdict(list)
    in_out=in_out_dict(graph)
    g_inds = set(graph.keys())
    chains = []

    #determine block-variables
    while(len(g_inds) != 0):
        chain_i = find_max_chain(list(g_inds)[0],graph, in_out)
        g_inds = set(g_inds) - set(chain_i)
        chains.append(chain_i)

    res=[[all(np.isin(chains[j], i)) for i in chains] for j in range(len(chains))]
    resnew=[]
    for i in range(len(res)):
        if (res[i].count(True) == 1):
            resnew.append(chains[i])

    chains = resnew
    
    #determine block-graph
    block_graph = defaultdict(list)
    block_nodes = [i[0] for i in chains]
    for i in range(len(chains)):
        block_graph[block_nodes[i]] = graph[chains[i][-1]]
        
    return in_out, chains, block_graph


#permutation generation
def comp(setlist):
    set_lengths = [len(i) for i in setlist]
    max_len = np.argmax(set_lengths)
    del set_lengths[max_len]
    next_max = np.argmax(set_lengths)
    return setlist[max_len]-setlist[next_max]
        
#speculative & needs verification for neighbors belonging to 3 or more levels
def nodes_above_below(nodes, levels,filtered_graph_i):
    nodelevels=invert_dict_of_lists(levels) #node/level k/v
    #invert edges
    rg=add_missing_nodes(reverse_graph(filtered_graph_i))
    levelsrg = level_sort(rg) #get its level sort
    nodelevelsrg=invert_dict_of_lists(levelsrg)

    #at and above
    l1ls=[set.union(*[set(i) for i in list(levels.values())[j-1:]]) for j in range(1,np.max(list(levels.keys()))+2)]
    l1l=[len(i) for i in l1ls]
    #at and below
    l2ls=[set.union(*[set(i) for i in list(levelsrg.values())[j-1:]]) for j in range(1,np.max(list(levelsrg.keys()))+2)]
    l2l=[len(i) for i in l2ls]

    above_below={}
    for i in nodes:
        nnodes = filtered_graph_i[i]
        nnodesr = rg[i]
        if(nnodes):
            adj_level = np.min([nodelevels[j] for j in nnodes])
            val = l1l[adj_level]#how many at and above given index
            cumsums=np.unique([l1ls[i] for i in [nodelevels[j] for j in nnodes]])
            if (len(cumsums) >= 2):#individual nodes can belong to multiple levels, distorting the count
                val=len(comp(cumsums))
        else:
            val = 0

        if(nnodesr):
            adj_level = np.min([nodelevelsrg[j] for j in nnodesr])
            valb=l2l[adj_level]#how many at and below given index
            cumsums=np.unique([l2ls[i] for i in [nodelevelsrg[j] for j in nnodesr]])
            if (len(cumsums) >= 2):
                valb=len(comp(cumsums))
        else:
            valb = 0
            
        above_below[i] = [valb, val]
    return above_below

def lstlist2(lst):
    return [list([*i]) for i in lst]

def filter_duplicates(lists):
    return [lst for lst in lists if all(x != y for x, y in zip(lst, lst[1:]))]

def products2(lst):
    blocks = []
    for v in lst:
        blocks.append(lstlist2(filter_duplicates(list(itertools.product(*v)))))
    return blocks

#produce permutations for a given disconnected component
def get_perms(n, nodes, above_below, levels):
    legal_values = [nodes[0+i[0]:n-i[1]] for i in above_below.values()]
    #deals with node sets that aren't 0-based and leaves them in the same order.
    node_map = dict(zip(nodes,np.arange(len(nodes))))#group_nodes[dcompi]
    block_variables=[[list(legal_values[node_map[j]]) for j in i] for i in list(levels.values())]
    ex2=products2(block_variables)
    res = [[elem for s in i for elem in s] for i in list(itertools.product(*ex2))] #these are products of all possibilities
    res = [i for i in res if len(set(i))==len(i)] #these are filtered plausible products
    return res

#2 macro methods for user convenience
def fixed_perms(group_pos, group_nodes):
    #topo sort is used in the transitive reduction
    filtered_graph = [add_missing_nodes(transitive_reduction(j)) for j in [edges_to_adjacency_list(i) for i in group_pos]]
    
    graph_chains = []
    for i in range(len(filtered_graph)):
        in_out, chains, block_graph = contract_graph(filtered_graph[i])
        graph_chains.append(chains)
        
    dc_perms=[]
    for dcompi in range(len(filtered_graph)):
        #print(dcompi)
        nodesi = group_nodes[dcompi]
        n = len(nodesi)
        levels = level_sort(filtered_graph[dcompi])
    
        above_below = nodes_above_below(nodesi,levels, filtered_graph[dcompi])
    
        res = get_perms(n, nodesi,above_below, levels)
        dc_perms.append(res)
    
    legal_values = [group_nodes[0+i[0]:n-i[1]] for i in above_below.values()]
        
    #then product across
    canonical_perms = [[elem for s in i for elem in s] for i in list(itertools.product(*dc_perms))]
    return canonical_perms

def comb_labelings(nodes, group_nodes):
    #fencepost
    products =itertools.combinations(nodes, len(group_nodes[0]))
    products = [list(j) for j in products]
    
    for i in range(1,len(group_nodes)-1):
        subsets=itertools.combinations(nodes, len(group_nodes[i]))
        subsets = [list(j) for j in subsets]
        #sorting here unnecessary
        possible_labels=[[list(np.sort(list(set(j) - set(products[k])))) for j in subsets if len(list(set(j) - set(products[k]))) == len(group_nodes[i])] for k in range(len(products))]
        factors = [[[products[i]], possible_labels[i]] for i in range(len(products))]
        #need to flatten, can reduce these to one line
        products = [[[elem for s in i for elem in s] for i in list(itertools.product(*factors[j]))] for j in range(len(factors))]
        products = [elem for s in products for elem in s]
    
    #fencepost (saves time of taking combinations)
    remainder = [list(np.sort(list(set(nodes) - set(i)))) for i in products]
    if (remainder != [[]]):
        composite = [products[i]+remainder[i] for i in range(len(products))]
    else:
        composite = products
        
    return composite

def get_phi(nodes, group_pos, group_nodes, prints=False):
    #canonical perms is the list of permutations allowed for the fixed node labeling 0...n.
    canonical_perms = fixed_perms(group_pos, group_nodes)
    #composite is the list of possible node labelings (up to combination),
    combination_labels = comb_labelings(nodes, group_nodes)
    if (prints):
        print("fixed-order (1..n) permutations: ", len(canonical_perms), canonical_perms[0:3], "\n")
        print("node labels (up to combination): ", len(combination_labels), combination_labels[0:3])
    
    #permutation list phi is the canonical perms under all node labelings up to combination
    phi = []
    for i in combination_labels:
        dictt = dict(zip(nodes, i))
        conversions =[[dictt[j] for j in k] for k in canonical_perms]
        phi.append(conversions)
    phi = [np.array(j) for j in [elem for s in phi for elem in s]]
    return phi

def filtrate(perm):
    sd=[]
    for i in range(len(perm)-1):
        dif = perm[i+1]-perm[i]
        sd.append(int((dif/np.abs(dif) + 1) / 2))
    return tuple(sd)

def genmat(phi, n, nodes):
    phi_inv = [perm_inverse(i) for i in phi]
    A = [0]*(n*n)
    for i in range(n):
        for j in range(n):
            A[n*i + j] = filtrate(tuple(nodes[phi_inv[i]][phi[j]])) #remember order of application
            #if (!np.isin(uniques,A[n*i + j])[0]):
    return A

def symsub(A,n,nodes):
    Aunq = list(set(A))
    s=sp.symbols("x_0:{}".format(len(Aunq)))
    
    if tuple([1]*(nodes-1)) in Aunq:
        ind = Aunq.index(tuple([1]*(nodes-1)))
        Aunq[ind], Aunq[0] = Aunq[0], Aunq[ind]
        
    subs = {k: v for k, v in zip(Aunq, s)}
    A = sp.Matrix([subs[i] for i in A]).reshape(n,n)
    return A


import unittest
from zlepy import eigs
import sympy as sp
import numpy as np
from tests.testing_utils import validate_eigs
import sys

np.set_printoptions(threshold=sys.maxsize)

N = 3
SYMBOLS_03 = sp.symbols('x_0:3')

A_03 = sp.Matrix(
        [
[-4*SYMBOLS_03[0], 0, 0], [0,-4*SYMBOLS_03[2],0], [0,0,-SYMBOLS_03[0]-2*SYMBOLS_03[1]-3*SYMBOLS_03[2]]
        ]
)
A_03s = sp.Matrix(
        [
[-44*SYMBOLS_03[2], 0, 0], [0,-43*SYMBOLS_03[2],0], [0,0,-6*SYMBOLS_03[0]+55*SYMBOLS_03[1]-6*SYMBOLS_03[2]]
        ]
)

class TestEigCoeffs(unittest.TestCase):
    
    def test_eig_with_max_base_case(self):
        result = eigs(A_03, SYMBOLS_03, batch_size=3, parallel=False, staggered=False)
        self.assertTrue(validate_eigs(result, A_03, SYMBOLS_03))

    def test_eig_with_max_staggered(self):
        result = eigs(A_03s, SYMBOLS_03, batch_size=3, parallel=False, staggered=True)
        self.assertTrue(validate_eigs(result, A_03s, SYMBOLS_03))

    def test_eig_parallel_base_case(self):
        result = eigs(A_03, SYMBOLS_03, batch_size=3, parallel=True, staggered=False)
        self.assertTrue(validate_eigs(result, A_03, SYMBOLS_03))

    def test_eig_parallel_staggered(self):
        result = eigs(A_03s, SYMBOLS_03, batch_size=3, parallel=True, staggered=True)
        self.assertTrue(validate_eigs(result, A_03s, SYMBOLS_03))



# SPDX-FileCopyrightText: 2024-present Abraham Miller <abraham.r.m@gmail.com>
#
# SPDX-License-Identifier: MIT

import unittest
from zlepy import eigs
from sympy import Symbol, Matrix

class TestEig(unittest.TestCase):
    def test_eig(self):
        matrix = Matrix([[Symbol('a'), Symbol('b')], [Symbol('b'), Symbol('a')]])
        expected = Matrix([[Symbol('a'), -Symbol('b')], [Symbol('a'), Symbol('b')]])
        result = eigs(matrix, [Symbol('a'), Symbol('b')], parallel=False)
        self.assertEqual(result, expected, 'The eig function should return the correct coefficients for each eigenvalue.')

class TestParallelEig(unittest.TestCase):
    def test_parallel_eig(self):
        matrix = Matrix([[Symbol('a'), Symbol('b')], [Symbol('b'), Symbol('a')]])
        expected = Matrix([[Symbol('a'), -Symbol('b')], [Symbol('a'), Symbol('b')]])
        result = eigs(matrix, [Symbol('a'), Symbol('b')], parallel=True)
        self.assertEqual(result, expected, 'The eig function should return the correct coefficients for each eigenvalue.')



import numpy as np
from collections import defaultdict, deque
import itertools
import sympy as sp

def get_po(f):
    factors = f[:]
    if (len(factors) == 1):
        factors.append(0)
    finlist = [[*[(i, i+j) for i in range(int(np.sum(factors[0:l-1])+1),int(np.sum(factors[0:l])-1)) for j in range(2,np.min([np.sum(factors[0:l])-i+1,4]))],*[(np.sum(factors[0:l]) -k, j) for k in range(0,np.min([factors[l-1],2])) for j in range(np.sum(factors[0:l])+1,np.sum(factors[0:l])+np.min([3,factors[l]+1]))]] for l in range(1,len(factors))]
    return [elem for s in [elem for s in finlist for elem in s] for elem in s]
    
def rebuildlists(k):
    res=[]
    lsts=[[1],[0]]
    if ((k-1) == 0):
        return lsts
        
    for i in range(k-1):
        res=[]
        for j in lsts:
            if j[-1]==1:
                res.append(j+[0])
                res.append(j+[1])
            else:
                res.append(j+[1])
        lsts=res
    return res

def adfunc(fac):
    out = [rebuildlists(i-1) for i in fac]
    [[j.append(1) for j in i] for i in out[0:len(out)-1]]
    out=[[elem for s in i for elem in s] for i in list(itertools.product(*out))]
    return out
    
def set_conv(lst):
    if (len(lst) == 0):
        return set()
    else:
        return set.union(*[set(i) for i in lst])

def list_set_mix(lst1, lst2):
    if (len(lst1) == 0 and len(lst2) == 0):
        return []
    elif (len(lst1) == 0):
        return [elem for s in [list(set(i)) for i in lst2] for elem in s]
    elif (len(lst2) == 0):
        return [elem for s in [list(set(i)) for i in lst1] for elem in s]
    else:
        return list(set.union(*[set(i) for i in lst2], *[set(i) for i in lst1]))
        
def lstlist2(lst):
    return [list([*i]) for i in lst]

def tupsub(tup):
    return np.abs(tup[0]-tup[1])

def s_alg(s1,s2, nodes):
    res = s1.copy()
    z1=[i for i in range(nodes-1) if s1[i]==0]
    z2=[i for i in range(nodes-1) if s2[i]==0]
    prodinds=lstlist2(list(itertools.product(*[z1,z2])))
    difs=[tupsub(i) for i in prodinds]
    zdif=[prodinds[i] for i in range(len(prodinds)) if difs[i]==0]
    odif=[prodinds[i] for i in range(len(prodinds)) if difs[i]==1]
    unqs = list_set_mix(odif, zdif)
    for i in unqs:
        res[i]=1 
    for i in list(set(z2) - set_conv(zdif)):
        res[i]=0
    return res

def fgenmat(sdx,n,nodes):
    #sdx is bigger than n
    A = [0]*(n*n)
    for i in range(n):
        for j in range(n):
            A[n*i + j] = tuple(s_alg(sdx[i], sdx[j], nodes))
    return A
    
def fsymsub(A,n,nodes, adstrs):
    s=sp.symbols("x_0:{}".format(len(adstrs)))
    if tuple([1]*(nodes-1)) in adstrs:
        ind = adstrs.index(tuple([1]*(nodes-1)))
        adstrs[ind], adstrs[0] = adstrs[0], adstrs[ind]
    subs = {k: v for k, v in zip(adstrs, s)}
    A = sp.Matrix([subs[i] for i in A]).reshape(n,n)
    return A


import unittest
import numpy as np
from zlepy.gen import (
    argsort,
    perm_inverse,
    po_groups,
    edges_to_adjacency_list,
    merge_components,
    add_missing_nodes,
    topological_sort,
    longest_path_dag,
    transitive_reduction,
    level_sort,
    invert_dict_of_lists,
    reverse_graph,
    compute_degrees,
    find_max_chain,
    in_out_dict,
    contract_graph
)

class TestGen(unittest.TestCase):

    def test_argsort(self):
        seq = [3, 1, 4, 1, 5, 9, 2, 6, 5, 3, 5]
        result = argsort(seq)
        self.assertEqual(result, [1, 3, 6, 0, 9, 2, 4, 8, 10, 7, 5])

    def test_perm_inverse(self):
        perm = np.array([2, 0, 1])
        result = perm_inverse(perm)
        np.testing.assert_array_equal(result, np.array([1, 2, 0]))

    def test_po_groups(self):
        po = np.array([0, 1, 1, 2, 2, 3, 3, 4])
        result = po_groups(po)
        expected = np.array([[0, 1, 1, 2, 2, 3, 3, 4]])
        self.assertTrue(np.array_equal(result, expected))

    def test_edges_to_adjacency_list(self):
        edges = [0, 1, 1, 2, 2, 3]
        result = edges_to_adjacency_list(edges)
        expected = {0: [1], 1: [2], 2: [3], 3: []}
        self.assertEqual(result, expected)

    def test_merge_components(self):
        comps = [{0: [1, 2]}, {3: [4, 5]}]
        result = merge_components(comps)
        expected = {0: [1, 2], 3: [4, 5]}
        self.assertEqual(result, expected)

    def test_add_missing_nodes(self):
        edge_list = {0: [1, 2], 1: [3]}
        result = add_missing_nodes(edge_list)
        expected = {0: [1, 2], 1: [3], 2: [], 3: []}
        self.assertEqual(result, expected)

    def test_topological_sort(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = topological_sort(graph)
        self.assertEqual(result, [0, 1, 2, 3])

    def test_longest_path_dag(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = longest_path_dag(graph)
        expected = {0: 0, 1: 1, 2: 1, 3: 2}
        self.assertEqual(result, expected)

    def test_transitive_reduction(self):
        graph = {0: [1, 2, 3], 1: [3], 2: [3], 3: []}
        result = transitive_reduction(graph)
        expected = {0: [1, 2], 1: [3], 2: [3], 3: []}
        self.assertEqual(result, expected)

    def test_level_sort(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = level_sort(graph)
        expected = {0: [0], 1: [1, 2], 2: [3]}
        self.assertEqual(result, expected)

    def test_invert_dict_of_lists(self):
        d = {0: [1, 2], 1: [3, 4]}
        result = invert_dict_of_lists(d)
        expected = {1: 0, 2: 0, 3: 1, 4: 1}
        self.assertEqual(result, expected)

    def test_reverse_graph(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = reverse_graph(graph)
        expected = {1: [0], 2: [0], 3: [1, 2]}
        self.assertEqual(result, expected)

    def test_compute_degrees(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = compute_degrees(graph, 1)
        self.assertEqual(result, [1, 1])

    def test_find_max_chain(self):
        graph = {0: [1], 1: [2], 2: [3], 3: []}
        in_out = {0: [0, 1], 1: [1, 1], 2: [1, 1], 3: [1, 0]}
        result = find_max_chain(0, graph, in_out)
        self.assertEqual(result, [0, 1, 2, 3])

    def test_in_out_dict(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = in_out_dict(graph)
        expected = {0: [0, 2], 1: [1, 1], 2: [1, 1], 3: [2, 0]}
        self.assertEqual(result, expected)

    def test_contract_graph(self):
        graph = {0: [1, 2], 1: [3], 2: [3], 3: []}
        result = contract_graph(graph)
        self.assertEqual(len(result), 3)  # Check if it returns 3 items (in_out, chains, block_graph)
        self.assertIsInstance(result[0], dict)
        self.assertIsInstance(result[1], list)
        self.assertIsInstance(result[2], dict)

if __name__ == '__main__':
    unittest.main()


# SPDX-FileCopyrightText: 2024-present Abraham Miller <abraham.r.m@gmail.com>
#
# SPDX-License-Identifier: MIT

import numpy as np
import sympy as sp
from multiprocessing import Pool, cpu_count
from mpmath import mp

def eigs(A, symbols, parallel=True, batch_size=None, pool_size=cpu_count(), staggered=False, base=10):
    """
    Symbolic Eigenvalue Solver (for integer linear eigenvalues).
    Calculates symbolic integer linear eigenvalues for a given compatible matrix.
    Parameters:
       A (2D Matrix): The input symbolic matrix for which eigenvalues are calculated.
                      Matrix must be square and be of a form which can have integer linear eigenvalues.
       symbols (list): The symbols used in the input matrix. Must be of length n (the number of rows/columns).
       parallel (bool): If True, use multiprocessing to calculate the eigenvalues in parallel. Defaults to True.
       batch_size (int): The number of eigenvectors to calculate in each batch. If None, this will be set automatically.
       pool_size (int): The number of processes to use in the multiprocessing pool. Defaults to the number of CPU cores available.
       staggered (bool): If True, use staggered precision to account for larger coefficients. Defaults to False.
       base (int): The base to use for calculations. Defaults to 10.
    Returns:
       (Matrix) The symbolic eigenvalues as linear combinations of the input symbols, 
                each represented by a list of integer coefficients on the symbols from the input matrix.
    """
    base = mp.mpf(base)

    if parallel:
        print(f"Calculating eigenvalues in parallel, using {pool_size} processes.")
        return _eigs_parallel(A, symbols, batch_size, pool_size, staggered, base)
    else:
        return _eigs_synchronous(A, symbols, batch_size, staggered, base)

def _eigs_parallel(A, symbols, batch_size, pool_size, staggered, base):
    """
    Calculates symbolic integer linear eigenvalues for a given compatible matrix in parallel.
    """
    n = np.shape(A)[0]
    assert n == np.shape(A)[1], "Matrix must be square."
    assert len(symbols) == n, "The number of symbols must be equal to the number of rows/columns in the matrix."

    if batch_size is None:
        batch_size = int(np.ceil(n/pool_size))
    
    digits = 2*batch_size if staggered else batch_size
    mp.dps = digits + 10

    num_batches = int(np.ceil(n/batch_size))

    extraction_params = _set_extraction_params(base, digits, staggered)
    dynamics = _set_dynamics(base, n, digits, staggered)
    
    results_by_batch = _calculate_eigenvalue_batches_in_parallel(A, n, batch_size, num_batches, symbols, extraction_params, dynamics, pool_size, staggered, base)
    eigens = _concatenate_batches(results_by_batch, n, batch_size, num_batches)

    annotated_eigens = eigens*np.array([symbols]*n)
    return sp.Matrix(annotated_eigens)

def _eigs_synchronous(A, symbols, batch_size=None, staggered=False, base=10):
    """
    Calculates symbolic integer linear eigenvalues for a given compatible matrix, without multiprocessing.
    """
    n = np.shape(A)[0]
    assert n == np.shape(A)[1], "Matrix must be square."
    assert len(symbols) == n, "The number of symbols must be equal to the number of rows/columns in the matrix."

    if batch_size is None:
        batch_size = n
    
    digits = 2*batch_size if staggered else batch_size
    mp.dps = digits + 10

    num_batches = int(np.ceil(n/batch_size))

    extraction_params = _set_extraction_params(base, digits, staggered)
    dynamics = _set_dynamics(base, n, digits, staggered)
    
    results_by_batch = _calculate_eigenvalue_batches_synchronously(A, n, batch_size, digits, num_batches, symbols, extraction_params, dynamics, staggered, base)
    eigens = _concatenate_batches(results_by_batch, n, batch_size, num_batches)

    annotated_eigens = eigens*np.array([symbols]*n)
    return sp.Matrix(annotated_eigens)

def _set_extraction_params(base, digits, staggered):
    midpoint_value = int(mp.floor(base/2))-1
    if staggered:
        midpoint = sum([mp.mpf(midpoint_value * base ** i) for i in range(-int(np.floor(digits/2)),int(np.ceil(digits/2)))])
    else:
        midpoint = mp.nsum(lambda i: midpoint_value*base**i, [-int(np.floor(digits/2)),int(np.ceil(digits/2)-1)])
    return {
        'midpoint_value': midpoint_value,
        'midpoint': midpoint
    }

def _set_dynamics(base, n, digits, staggered):
    if staggered:
        indicators = [mp.mpf(base ** i) for i in range(-int(np.floor(digits/2)),int(np.ceil(digits/2))-1,2)][::-1]
    else:
        indicators = [mp.mpf(base**i) for i in range(-int(mp.floor(digits/2)),int(mp.ceil(digits/2)))][::-1]

    reals = list(range(n))
    reals_duplicate_check = [1,1]
    while (len(reals_duplicate_check) != len(list(set(reals_duplicate_check)))):
        reals = list(mp.randmatrix(1,n))
        reals_duplicate_check = [mp.nstr(i.real, int(np.floor(digits/2))) for i in reals]
        
    return {'real_values': reals, 'indicator_values': indicators}

def _calculate_eigenvalue_batches_in_parallel(A, n, batch_size, num_batches, symbols, extraction_params, dynamics, pool_size, staggered, base):
    with Pool(processes=pool_size) as pool:
        results = pool.starmap(_calculate_eigenvalues_for_batch, 
                               [(A, n, batch_index, batch_size, batch_size*(2 if staggered else 1), symbols, extraction_params, dynamics, staggered, base) 
                                for batch_index in range(num_batches)])
    return results

def _calculate_eigenvalue_batches_synchronously(A, n, batch_size, digits, num_batches, symbols, extraction_params, dynamics, staggered, base):
    results = [None] * num_batches
    for batch_index in range(num_batches):
        results[batch_index] = _calculate_eigenvalues_for_batch(A, n, batch_index, batch_size, digits, symbols, extraction_params, dynamics, staggered, base)
    return results

def _calculate_eigenvalues_for_batch(A, n, batch_index, batch_size, digits, symbols, extraction_params, dynamics, staggered, base):
    mp.dps = digits + 10
    Ai = _pseudo_symbolic_matrix_segment(A, n, batch_index, batch_size, symbols, dynamics)
    eigenvalues, eigenvectors = mp.eig(Ai)
    eigenvalues, eigenvectors = mp.eig_sort(eigenvalues, eigenvectors) 
    return _extract_symbolic_eigenvalues(eigenvalues, batch_size, digits, extraction_params['midpoint'], extraction_params['midpoint_value'], staggered, base)

def _extract_symbolic_eigenvalues(eigenvalues, batch_size, digits, midpoint, midpoint_value, staggered, base):
    with mp.workdps(digits):
        midpoint_normalized_eigens = [mp.im(i)+midpoint for i in eigenvalues]
        mne_strings = [get_strings(i, digits, midpoint, midpoint_value, base) for i in midpoint_normalized_eigens]
   
    coeffs = [_extract_coef_list(eigenvalue_string, midpoint_value) for eigenvalue_string in mne_strings]
    if staggered:
        coeffs = [[np.sum([i*j for i,j in zip([base,1], x)]) for x in [ci[i:i+2] for i in range(0, digits,2)]] for ci in coeffs]
    return coeffs

def get_strings(ev, digits, midpoint, midpoint_value, base):
    zero_fill = filler_digits(ev, digits, base)

    with mp.workdps(digits):
        esize = digits - zero_fill
        mne_string = mp.nstr(ev, esize, strip_zeros=False, min_fixed=-mp.inf, max_fixed=mp.inf)
        mne_string = '0'*(zero_fill) + mne_string 
        return mne_string

def filler_digits(e, digits, base):
    if (np.abs(e) / (base ** int(np.ceil(digits/2)-1)) >= 1):
        return 0
    j = 2
    while (np.abs(e) / (base ** int(np.ceil(digits/2)-j)) < 1):
        j += 1
    return min(int(np.ceil(digits/2)-1), (j-1))
        
def _extract_coef_list(eigenvalue_string, midpoint_value):
    val = list(eigenvalue_string)
    del val[val.index('.')]
    return [int(i) - midpoint_value for i in val]

def _pseudo_symbolic_matrix_segment(A, n, batch_index, batch_size, symbols, dynamics):
    real_values = dynamics['real_values'].copy()
    indicator_values = dynamics['indicator_values']

    for i in range(batch_index*batch_size, np.min([n,(batch_index+1)*batch_size])):
        real_values[i] = mp.mpc(real_values[i].real, indicator_values[i-batch_index*batch_size])

    subs = {k: v for k, v in zip(symbols, real_values)}
    Af = A.subs(subs)
    return mp.matrix(Af)

def _concatenate_batches(results_by_batch, n, batch_size, num_batches):
    eigens_by_batch = np.array([np.zeros((n,batch_size), dtype=int)]*(num_batches))
    for batch_index in range(num_batches):
        eigens_by_batch[batch_index] = results_by_batch[batch_index]

    return np.array([np.hstack([eigens_by_batch[i,j] for i in range(num_batches)]) for j in range(n)])[0:n,0:n]



import numpy as np
import sympy as sp

def validate_eigs(symsol,A,symbols):
    n = A.shape[0]
    for i in range(2):
        rvals=np.random.rand(n)*2-1
        At = np.array(A.subs({k: v for k, v in zip(symbols, rvals)})).astype(np.float64)
        w,v = numeric_eig(At)
        sw = np.array(sp.Matrix(symsol).subs({k: v for k, v in zip(symbols, rvals)})).astype(np.float64)
        sw = np.sort([np.sum(sw[j,:]) for j in range(n)])
        if (np.array_equal(np.round(w,3),np.round(sw,3))==False):
            print(w,'\n\n',sw)
            return False
    return True

def numeric_eig(A):
    eigenvalues, eigenvectors = np.linalg.eig([A])
    eigenvalues = reorder(eigenvalues, eigenvectors)[0][0]
    return eigenvalues, eigenvectors

def reorder(
    eigenvalues, eigenvectors
) -> tuple:
    indices_sort_all = np.argsort(eigenvalues.real)
    for i in range(len(eigenvalues)):
        indices_sort = indices_sort_all[i]

        eigenvalues[i] = eigenvalues[i][indices_sort]
        eigenvectors[i] = eigenvectors[i][:, indices_sort]
    return eigenvalues, eigenvectors
